//SN76489 sound chip emulator v1.02
//by Shiru (shiru@mail.ru) 07'09

package vgmplayer
{
import flash.utils.*;

public class SN76489
{
    private const volumeTable:Array=[
        .25,.2442,.1940,.1541,.1224,.0972,.0772,.0613,.0487,.0386,.0307,.0244,.0193,.0154,.0122,0,
        -.25,-.2442,-.1940,-.1541,-.1224,-.0972,-.0772,-.0613,-.0487,-.0386,-.0307,-.0244,-.0193,-.0154,-.0122,0,
        .25,.2442,.1940,.1541,.1224,.0972,.0772,.0613,.0487,.0386,.0307,.0244,.0193,.0154,.0122,0,
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    ];
 
    private var volA:uint;
    private var volB:uint;
    private var volC:uint;
    private var volD:uint;
    private var divA:int;
    private var divB:int;
    private var divC:int;
    private var divD:int;
    private var cntA:int;
    private var cntB:int;
    private var cntC:int;
    private var cntD:int;
    private var outA:Number;
    private var outB:Number;
    private var outC:Number;
    private var outD:Number;
    private var noiseLFSR:uint;
    private var noiseTap:uint;
    private var latchedChan:uint;
    private var latchedVolume:Boolean;
 
    private var ticksPerSample:Number;
    private var ticksCount:Number;

    public function SN76489():void
    {
        clock(3500000);
        reset();
    }
 
    public function clock(f:Number):void
    {
        ticksPerSample=f/16/44100;
    }
 
    public function reset():void
    {
        volA=15;
        volB=15;
        volC=15;
        volD=15;
        outA=0;
        outB=0;
        outC=0;
        outD=0;
        latchedChan=0;
        latchedVolume=false;
        noiseLFSR=0x8000;
        ticksCount=ticksPerSample;
    }

    public function getDivByNumber(chan:uint):uint
    {
        switch(chan)
        {
        case 0: return divA;
        case 1: return divB;
        case 2: return divC;
        case 3: return divD;
        }
        return 0;
    }

    public function setDivByNumber(chan:uint,div:uint):void
    {
        switch(chan)
        {
        case 0: divA=div; break;
        case 1: divB=div; break;
        case 2: divC=div; break;
        case 3: divD=div; break;
        }
    }

    public function getVolByNumber(chan:uint):uint
    {
        switch(chan)
        {
        case 0: return volA;
        case 1: return volB;
        case 2: return volC;
        case 3: return volD;
        }
        return 0;
    }

    public function setVolByNumber(chan:uint,vol:uint):void
    {
        switch(chan)
        {
        case 0: volA=vol; break;
        case 1: volB=vol; break;
        case 2: volC=vol; break;
        case 3: volD=vol; break;
        }
    }
 
    public function write(val:int):void
    {
        var chan:int;
        var cdiv:int;
        var div:int;
        var vol:int;
 
        if(val&128)
        {
            chan=(val>>5)&3;
            cdiv=(getDivByNumber(chan)&0xfff0)|(val&15);
 
            latchedChan=chan;
            latchedVolume=val&16?true:false;
        }
        else
        {
            chan=latchedChan;
            cdiv=(getDivByNumber(chan)&15)|((val&63)<<4);
        }

        if(latchedVolume)
        {
            setVolByNumber(chan,(getVolByNumber(chan)&16)|(val&15));
        }
        else
        {
            setDivByNumber(chan,cdiv);
            if(chan==3)
            {
                noiseTap=((cdiv>>2)&1)?9:1;
                noiseLFSR=0x8000;
            }
        }
    }
 
    public function render(buf:ByteArray,len:uint):void
    {
        var i:uint;
        var cdiv:uint,tap:uint;
        var out:Number;
 
        for(i=0;i<len;i+=1)
        {
            while(ticksCount>0)
            {
                cntA-=1;
                if(cntA<0)
                {
                    if(divA>1)
                    {
                        volA^=16;
                        outA=Number(volumeTable[volA]);
                    }
                    cntA=divA;
                }
 
                cntB-=1;
                if(cntB<0)
                {
                    if(divB>1)
                    {
                        volB^=16;
                        outB=Number(volumeTable[volB]);
                    }
                    cntB=divB;
                }
 
                cntC-=1;
                if(cntC<0)
                {
                    if(divC>1)
                    {
                        volC^=16;
                        outC=Number(volumeTable[volC]);
                    }
                    cntC=divC;
                }
 
                cntD-=1;
                if(cntD<0)
                {
                    cdiv=divD&3;
                    if(cdiv<3) cntD=0x10<<cdiv; else cntD=divC<<1;
 
                    if(noiseTap==9)
                    {
                        tap=noiseLFSR&noiseTap;
                        tap^=tap>>8;
                        tap^=tap>>4;
                        tap^=tap>>2;
                        tap^=tap>>1;
                        tap&=1;
                    }
                    else
                    {
                        tap=noiseLFSR&1;
                    }
 
                    noiseLFSR=(noiseLFSR>>1)|(tap<<15);
                    volD=(volD&15)|((noiseLFSR&1^1)<<4);
                    outD=Number(volumeTable[volD]);
                }
 
                ticksCount-=1;
            }

            ticksCount+=ticksPerSample;
            out=outA+outB+outC+outD;

            buf.writeFloat(out);
            buf.writeFloat(out);
        }
    }
}
}